// Robert Pciennik G1ISI

#include <boost/program_options.hpp>

#include <iostream>
#include <fstream>
#include <iterator>
#include <vector>

namespace po = boost::program_options;

// vector<T> -> "T1 T2 T3..."

template<class T>
std::ostream& operator << ( std::ostream& os_, const std::vector<T>& vec_ )
{
    copy(vec_.begin(), vec_.end(), std::ostream_iterator<T>(std::cout, " ")); 
    return os_;
}

int main( int argc, char* argv[] )
{
	const char* defStr = "Sample string";
	const int defInt = 13;
	const char* envPrefix = "TEST_";
	
	try
	{	
		// options_description stanowi kontener przechowujcy definiowane opcje
		// programu, ktre pniej mona oczytywa parsujc command-line, config
		// files, czy nawet dostpne zmienne rodowiska w ramach kontenera.
		// Opcjonalnym argumentem konstruktora options_description jest caption
		// sucy celom informacyjnym, gdy options_descriptor wypisywany jest
		// na wyjcie (przeciony operator <<).

		/**/ po::options_description argless( "Options without arguments" );

		// Dodawanie opcji do kontenera za pomoc metody add_options(), ktra
		// zwraca referencj na pomocniczy obiekt z przecionym
		// operatorem (), ktry z kolei zwraca referencj na samego siebie.
		// Pierwszym argumentem operatora () jest nazwa opcji, wraz z jej
		// opcjonalnym skrtem po przecinku, drugim - opis opcji.
		
		argless.add_options()
			( "help", "standard help" )
			( "long-help", "extended info" )
			( "version,v", "display version info and quit" )
		;

		// W powyszym przypadku, jeli dane opcje bd wykorzystywane jako
		// command-line options, wwczas 'version' bdzie rozpoznawana w linii
		// polecenia jako '--version' lub '-v'
		
		/**/ po::options_description withargs( "Options with arguments" );

		/**/ int int1var;
	
		// W celu zadeklarownia opcji przyjmujcych argumenty, naley
		// w operatorze () umieci jeszcze jeden argument wiadczcy o typie
		// danej opcji. Dany argument jest obiektem, uzyskiwanym przez wywoanie
		// konstruktora value< T >(), ktrego opcjonalnym argumentem moe by
		// wskanik na zmienn, ktra przy odczytaniu danej opcji bdzie
		// przechowywa jej warto.
		// Mona dodatkowo sprecyzowa warto domyln danej opcji za
		// porednictwem metody default_value() nowo utworzonego obiektu.
		
		withargs.add_options()
			( "int1", po::value< int >(&int1var) )
			( "int2def", po::value< int >()->default_value(defInt), "Has default value" )
			( "str1", po::value< std::string >()->default_value(defStr), "With default value as well" )
			( "str2vector", po::value< std::vector<std::string> >(), "Vector of strings" )
		;
		
		// W powyszym przypadku, warto opcji 'int1' po odczycie zachowana
		// bdzie pod zmienn int1var
		
		/**/ po::options_description notToBeListed( "Positional options" );
		
		notToBeListed.add_options()
			( "configs", po::value< std::vector<std::string> >(), "Vector of strings (2 first params)" )
			( "ints", po::value< std::vector<int> >(), "Vector of ints (2 next params)" )
		;
		
		// positional_options_description zawiera opcje zadeklarowane w pewnym
		// options_description (aby moliwy by ich odczyt) i suy odczytywaniu
		// wartoci opcji jako kolejnych argumentw podawanych w linii polecenia
		// bez precyzowania nazwy opcji jako argumentu
		
		/**/ po::positional_options_description pos;
		
		pos.add( "configs", 2 );
		pos.add( "ints", 2 );
		
		// W powyszym przypadku mona oczekiwa, e podanie w linii polecenia:
		// $./exe abc def 123
		// byoby jednoznaczne z:
		// $./exe --configs abc --configs def --ints 123
		
		// Zadeklarowanie opcji zczytywanych ze zmiennych rodowiska
		
		/**/ po::options_description envVars( "Environment variables" );

		envVars.add_options()
			( "envint", po::value< int >(), "Int from environment" )
			( "envstr", po::value< std::string >(), "String from environment" )
		;

		// options_description mona czy w bardziej oglne options_description
		// celem wykonywania parsowania na caym zbiorze opcji, bd ich
		// zgrupowanemu wywietleniu.
		// (Metoda add() klasy zwraca referencj na jej obiekt).
		
		/**/ po::options_description helpInfo( "Available options" );
		helpInfo.add( argless ).add( withargs );
		
		/**/ po::options_description extInfo( "All the options" );
		extInfo.add( argless ).add( withargs ).add( notToBeListed ).add( envVars );
		
		/**/ po::options_description parseCmdLine;
		parseCmdLine.add( argless ).add( withargs ).add( notToBeListed );
		
		// Dziki powyszemu zabiegowi:
		// * helpInfo zawiera te opcje, o ktrych informacje bd wywietlane
		//   przy uruchomieniu programu z parametrem/opcj --help
		// * extInfo zawiera informacje o opcjach wywietlane dla parametru
		//   --long-help, w tym informacje o opcjach, ktre nie powinny by
		//   w ten sposb wywietlane (notToBeListed, envVars)
		
		// variables_map zbliona jest do mapy zawierajca przyporzdkowanie
		// opcja -> warto, bd tylko wystpienie danej opcji dla opcji
		// bezargumentowych, gdy w wyniku parsowania, opcje takie zostay
		// zczytane
		
		/**/ po::variables_map vm;
		
		try
		{
			// Wszystkie metody pomocniczej klasy command_line_parser zwracaj
			// referencje do jej obiektu pozwalajc na wykonanie nastpujcych
			// operacji:
			// * parsowania linii polece pod ktem opcji dostpnych
			//   w parseCmdLine
			// * w tym uwzgldniajc positional options z 'pos'
			//
			// Efekt tej operacji ostatecznie zostaje przechowany w vm
			
			store( po::command_line_parser(argc,argv).options(parseCmdLine).positional(pos).run(), vm );
		}
		catch( std::exception& e_ )
		{
			std::cout << "While parsing command-line: " << e_.what() << "\n";
		}
		
		try
		{
			// Operacja parsowania zmiennych rodowiskowych o prefiksie
			// envPrefix zawarych w envVars
			//
			// Wszelkie zmienne rodowiska zgodne z prefiksem, lecz
			// nieuwzgldnione w envVars bed traktowane jako opcje nieznane
			// (tak samo, jak by to miao miejsce dla nieznanych paramterw
			// linii polecenia).
			// Dodatkowo nazwy opcji s przymusowo konwertowane do lowcase.
			//
			// Przykad:
			// 
			// Zmienna w rodowisku: MY_ENV_VARIABLE="My data"
			// envPrefix = "MY_"
			// envVars = { "Env_Variable" }
			//
			// W efekcie w vm znajdzie si: vm["env_variable"] == "My data"
			
			store( po::parse_environment(envVars,envPrefix), vm );
		}
		catch( std::exception& e_ )
		{
			std::cout << "While parsing environment: " << e_.what() << "\n";
		}

		// Wymagana operacja odwierzenia stanu variables_map		
		po::notify( vm );

		// Sprawdzanie wystpienia opcji, jest jednoznaczne ze sprawdzeniem jej
		// wystpowania w danej variables_map
		
		if( vm.count("configs") )
		{
			// Pliki sprecyzowane w opcji configs s kolejno parsowane.
			// Zapisane w nich przypisania opcja -> warto nie nadpisuj jednak
			// wartoi opcji zczytanych przy pomocy innych sposobw parsowania.
			//
			// Format pliku jest nastpujcy:
			// 
			// <option> = <value>
			// #<commented_option>
			//
			// Przykadowo, wczytanie:
			//
			// str1 = A new string
			//
			// spowoduje nadpisanie domylnej wartoci dla opcji str1, cho ta
			// moe zosta na powrot zmieniona chociby przekazujc warto
			// opcji w linii polecenia
			
			/**/ std::ifstream file;
			/**/ unsigned i;
			/**/ const std::vector< std::string >* configsPtr = &(vm["configs"].as< std::vector<std::string> >());

			for( i = 0; i <= (unsigned) configsPtr->size(); ++i )
				try
				{
					file.open( (*configsPtr)[i].c_str() );					
					store( parse_config_file(file,withargs), vm );

					// W tym przypadku zczytywane s tylko wartoci dla opcji
					// z 'withargs'					
				}
				catch( std::exception& e_ )
				{
					std::cout << "While parsing '" << (*configsPtr)[i] << "': " << e_.what() << "\n";
				}
		}

		// Wymagana operacja odwierzenia stanu variables_map		
		po::notify( vm );
		
		// Usage info
		if( vm.count("help") || vm.count("long-help") || argc == 1 )
		{
			std::cout << "./usage [OPTION <argument>] [configs{0}] [configs{1}] [ints{0}] [ints{1}]\n\n";
			
			// Zawarto options_description w postaci sformatowanej moe zosta
			// wysana na wyjcie
			
			if( vm.count("long-help") )
				std::cout << extInfo << "\n";
			else
				std::cout << helpInfo << "\n";
		}
		
		if( vm.count("version") )
		{
			std::cout << "Uses libBoost::program_options 1.33.1\n";
			return 0;
		}
		
		// Wypisanie wartoci danych opcji jeli te zostay zczytane, bd
		// posiaday wartoi domylne

		std::cout << "Options set:\n\n";
		
		// Dostpu do wartoi opcji z variables_map naley dokonywa przy
		// uyciu metody as() ustalajcej typ danej
		
		if( vm.count("int1") )
			std::cout << "  int1 == " << int1var << "\n";			
		if( vm.count("int2def") )
			std::cout << "  int2def == " << vm["int2def"].as< int >() << "\n";			
		if( vm.count("str1") )
			std::cout << "  str1 == " << vm["str1"].as< std::string >() << "\n";			
		if( vm.count("str2vector") )
			std::cout << "  str2vector == " << vm["str2vector"].as< std::vector<std::string> >() << "\n";			
		
		if( vm.count("configs") )
			std::cout << "  configs == " << vm["configs"].as< std::vector<std::string> >() << "\n";					
		if( vm.count("configs") )
			std::cout << "  ints == " << vm["ints"].as< std::vector<int> >() << "\n";					
		
		if( vm.count("envint") )
			std::cout << "  " << envPrefix << "ENVINT == " << vm["envint"].as< int >() << "\n";
		if( vm.count("envstr") )
			std::cout << "  " << envPrefix << "ENVSTR == " << vm["envstr"].as< std::string >() << "\n";
	}
	catch( std::exception& e_ )
	{
		std::cout << "While defining program options: " << e_.what() << "\n";
		return 1;
	}	
	
	return 0;
}
